import collections
import re


Token = collections.namedtuple("Token", ["character_index", "type", "value"])


class TokenType(object):
    IDENTIFIER = "identifier"
    SYMBOL = "symbol"
    WHITESPACE = "whitespace"
    STRING = "string"
    UNTERMINATED_STRING = "unterminated string"
    INTEGER = "integer"
    END = "end"
    


def regex_tokeniser(rules):
    rules = [(token_type, _to_regex(regex)) for token_type, regex in rules]
    rules.append(("unknown", re.compile(".")))
    
    def tokenise(value):
        tokens = []
        index = 0
        while index < len(value):
            for token_type, regex in rules:
                match = regex.match(value, index)
                if match is not None:
                    tokens.append(Token(index, token_type, match.group(0)))
                    index = match.end()
                    break
            else:
                # Should be impossible
                raise Exception("Remaining: " + value[index:])

        tokens.append(Token(index, TokenType.END, ""))

        return tokens

    return tokenise
    

def _to_regex(value):
    if hasattr(value, "match"):
        return value
    else:
        return re.compile(value)


_string_prefix = r"'(?:\\.|[^'])*"
_identifier_character = r"(?:[a-zA-Z\-_]|\\.)"

tokenise = regex_tokeniser([
    (TokenType.IDENTIFIER, _identifier_character + "(?:" + _identifier_character + "|[0-9])*"),
    (TokenType.SYMBOL, r":|>|=>|\^=|=|\(|\)|\[|\]|\||!|\."),
    (TokenType.WHITESPACE, r"\s+"),
    (TokenType.STRING, _string_prefix + "'"),
    (TokenType.UNTERMINATED_STRING, _string_prefix),
    (TokenType.INTEGER, "([0-9]+)"),
])



# TODO: check indices
# TODO: proper tests for unexpected tokens

from .errors import LineParseError


class TokenIterator(object):
    def __init__(self, tokens):
        self._tokens = tokens
        self._index = 0
    
    def peek_token_type(self):
        return self._tokens[self._index].type
    
    def next_value(self, token_type=None):
        return self._next(token_type).value
    
    def _next(self, token_type=None):
        token = self._tokens[self._index]
        if token_type is None or token.type == token_type:
            self._index += 1
            return token
        else:
            raise self._unexpected_token_type(token_type, token)
    
    def skip(self, token_type, token_value=None):
        token = self._tokens[self._index]
        if token.type == token_type and (token_value is None or token.value == token_value):
            self._index += 1
            return True
        else:
            raise self._unexpected_token_type(token_type, token)
    
    def try_skip(self, token_type, token_value=None):
        token = self._tokens[self._index]
        if token.type == token_type and (token_value is None or token.value == token_value):
            self._index += 1
            return True
        else:
            return False
    
    def try_skip_many(self, tokens):
        start = self._index
        for token_type, token_value in tokens:
            token = self._tokens[self._index]
            if not (token.type == token_type and (token_value is None or token.value == token_value)):
                self._index = start
                return False
            else:
                self._index += 1
        
        return True
    
    def _unexpected_token_type(self, token_type, token):
        raise LineParseError()
    



import cobble


class Node(object):
    pass


@cobble.data
class TextNode(Node):
    value = cobble.field()


@cobble.data
class Tag(object):
    tag_names = cobble.field()
    attributes = cobble.field()
    collapsible = cobble.field()
    separator = cobble.field()
    
    @property
    def tag_name(self):
        return self.tag_names[0]


@cobble.data
class Element(Node):
    tag = cobble.field()
    children = cobble.field()
    
    @property
    def tag_name(self):
        return self.tag.tag_name
        
    @property
    def tag_names(self):
        return self.tag.tag_names

    @property
    def attributes(self):
        return self.tag.attributes
        
    @property
    def collapsible(self):
        return self.tag.collapsible
        
    @property
    def separator(self):
        return self.tag.separator

    _VOID_TAG_NAMES = set(["br", "hr", "img"])

    def is_void(self):
        return not self.children and self.tag_name in self._VOID_TAG_NAMES


@cobble.visitable
class ForceWrite(Node):
    pass


NodeVisitor = cobble.visitor(Node)



import cobble

from . import html


def path(elements):
    return HtmlPath(elements)


def element(names, class_names=None, fresh=None, separator=None):
    if class_names is None:
        class_names = []
    if fresh is None:
        fresh = False
    if class_names:
        attributes = {"class": " ".join(class_names)}
    else:
        attributes = {}
    return HtmlPathElement(html.tag(
        tag_names=names,
        attributes=attributes,
        collapsible=not fresh,
        separator=separator,
    ))


@cobble.data
class HtmlPath(object):
    elements = cobble.field()
    
    def wrap(self, generate_nodes):
        nodes = generate_nodes()

        for element in reversed(self.elements):
            nodes = element.wrap_nodes(nodes)
        
        return nodes


@cobble.data
class HtmlPathElement(object):
    tag = cobble.field()

    def wrap(self, generate_nodes):
        return self.wrap_nodes(generate_nodes())

    def wrap_nodes(self, nodes):
        element = html.Element(self.tag, nodes)
        return [element]

empty = path([])


class ignore(object):
    @staticmethod
    def wrap(generate_nodes):
        return []



def uri_to_zip_entry_name(base, uri):
    if uri.startswith("/"):
        return uri[1:]
    else:
        return base + "/" + uri


def replace_fragment(uri, fragment):
    hash_index = uri.find("#")
    if hash_index != -1:
        uri = uri[:hash_index]
    return uri + "#" + fragment



import re

from .tokeniser import TokenType


def try_parse_class_name(tokens):
    if tokens.try_skip(TokenType.SYMBOL, "."):
        return parse_identifier(tokens)
    else:
        return None


def parse_identifier(tokens):
    return decode_escape_sequences(tokens.next_value(TokenType.IDENTIFIER))


def parse_string(tokens):
    return decode_escape_sequences(tokens.next_value(TokenType.STRING)[1:-1])


_ESCAPE_SEQUENCE_REGEX = re.compile(r"\\(.)")


def decode_escape_sequences(value):
    return _ESCAPE_SEQUENCE_REGEX.sub(_decode_escape_sequence, value)
    
    
def _decode_escape_sequence(match):
    code = match.group(1)
    if code == "n":
        return "\n"
    elif code == "r":
        return "\r"
    elif code == "t":
        return "\t"
    else:
        return code



import base64
import io
import os
import shutil
import subprocess
import tempfile

from . import html


def img_element(func):
    def convert_image(image):
        attributes = func(image).copy()
        if image.alt_text:
            attributes["alt"] = image.alt_text
            
        return [html.element("img", attributes)]
    
    return convert_image

# Undocumented, but retained for backwards-compatibility with 0.3.x
inline = img_element


@img_element
def data_uri(image):
    with image.open() as image_bytes:
        encoded_src = base64.b64encode(image_bytes.read()).decode("ascii")
    
    return {
        "src": "data:{0};base64,{1}".format(image.content_type, encoded_src)
    }


_wmf_extensions = {
    "image/x-wmf": ".wmf",
    "image/x-emf": ".emf",
}


def libreoffice_wmf_conversion(post_process=None):
    if post_process is None:
        post_process = lambda x: x
    
    def convert_image(image):
        wmf_extension = _wmf_extensions.get(image.content_type)
        if wmf_extension is None:
            return image
        else:
            temporary_directory = tempfile.mkdtemp()
            try:
                input_path = os.path.join(temporary_directory, "image" + wmf_extension)
                with io.open(input_path, "wb") as input_fileobj:
                    with image.open() as image_fileobj:
                        shutil.copyfileobj(image_fileobj, input_fileobj)
                
                output_path = os.path.join(temporary_directory, "image.png")
                subprocess.check_call([
                    "libreoffice",
                    "--headless",
                    "--convert-to",
                    "png",
                    input_path,
                    "--outdir",
                    temporary_directory,
                ])
                
                with io.open(output_path, "rb") as output_fileobj:
                    output = output_fileobj.read()
                
                def open_image():
                    return io.BytesIO(output)
                
                return post_process(image.copy(
                    content_type="image/png",
                    open=open_image,
                ))
            finally:
                shutil.rmtree(temporary_directory)
    
    return convert_image


def imagemagick_trim(image):
    command = ["convert", "-", "-trim", "-"]
    process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    try:
        with image.open() as image_fileobj:
            shutil.copyfileobj(image_fileobj, process.stdin)
        output, err_output = process.communicate()
    except:
        process.kill()
        process.wait()
        raise
        
    return_code = process.poll()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)
    else:
        def open_image():
            return io.BytesIO(output)
        
        return image.copy(open=open_image)



from . import documents


def paragraph(transform_paragraph):
    return element_of_type(documents.Paragraph, transform_paragraph)


def run(transform_run):
    return element_of_type(documents.Run, transform_run)


def element_of_type(element_type, transform):
    def transform_element(element):
        if isinstance(element, element_type):
            return transform(element)
        else:
            return element
    
    return _each_element(transform_element)


def _each_element(transform_element):
    def transform_element_and_children(element):
        if isinstance(element, documents.HasChildren):
            children = list(map(transform_element_and_children, element.children))
            element = element.copy(children=children)
        
        return transform_element(element)
    
    return transform_element_and_children


def get_descendants_of_type(element, element_type):
    return list(filter(
        lambda descendant: isinstance(descendant, element_type),
        get_descendants(element),
    ))


def get_descendants(element):
    descendants = []

    def visit(element):
        descendants.append(element)

    _visit_descendants(element, visit)

    return descendants


def _visit_descendants(element, visit):
    if isinstance(element, documents.HasChildren):
        for child in element.children:
            _visit_descendants(child, visit)
            visit(child)




from .errors import LineParseError
from .style_mapping_parser import parse_style_mapping
from .tokeniser import tokenise
from .token_iterator import TokenIterator
from ... import results


def read_style_mapping(string):
    try:
        tokens = tokenise(string)
        return results.success(parse_style_mapping(TokenIterator(tokens)))
    except LineParseError:
        warning = "Did not understand this style mapping, so ignored it: " + string
        return results.Result(None, [results.warning(warning)])
